import asyncio
import functools
import multiprocessing
import sys
import typing as t

from flama import compat

__all__ = ["is_async", "run", "run_task_group", "AsyncProcess"]

R = t.TypeVar("R", covariant=True)
P = compat.ParamSpec("P")  # PORT: Replace compat when stop supporting 3.9


def is_async(
    obj: t.Any,
) -> compat.TypeGuard[t.Callable[..., t.Awaitable[t.Any]]]:  # PORT: Replace compat when stop supporting 3.9
    """Check if given object is an async function, callable or partialised function.

    :param obj: Object to check.
    :return: True if it's an async function, callable or partialised function.
    """
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or asyncio.iscoroutinefunction(getattr(obj, "__call__"))


async def run(
    func: t.Union[t.Callable[P, R], t.Callable[P, t.Awaitable[R]]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    """Run a function either as asyncio awaiting it if it's an async function or running it in a thread if it's a
    sync function.

    :param func: Function to run.
    :param args: Positional arguments.
    :param kwargs: Keyword arguments.
    :return: Function returned value.
    """
    if is_async(func):
        return await func(*args, **kwargs)

    return t.cast(R, await asyncio.to_thread(func, *args, **kwargs))


if sys.version_info >= (3, 11):  # PORT: Remove when stop supporting 3.10 # pragma: no cover

    async def run_task_group(*tasks: t.Coroutine[t.Any, t.Any, t.Any]) -> list[asyncio.Task]:
        """Run a group of tasks.

        :param tasks: Tasks to run.
        :result: Finished tasks.
        """
        async with asyncio.TaskGroup() as task_group:
            return [task_group.create_task(task) for task in tasks]

else:  # pragma: no cover

    async def run_task_group(*tasks: t.Coroutine[t.Any, t.Any, t.Any]) -> list[asyncio.Task]:
        """Run a group of tasks.

        :param tasks: Tasks to run.
        :result: Finished tasks.
        """
        tasks_list = [asyncio.create_task(task) for task in tasks]
        await asyncio.wait(tasks_list)
        return tasks_list


class AsyncProcess(multiprocessing.Process):
    """Multiprocessing Process class whose target is an async function."""

    _target: t.Optional[t.Callable[..., t.Union[t.Any, t.Coroutine]]]
    _args: list[t.Any]
    _kwargs: dict[str, t.Any]

    def run(self) -> None:
        if self._target:
            result_or_task = self._target(*self._args, **self._kwargs)

            asyncio.run(result_or_task) if is_async(self._target) else result_or_task



import typing as t

import starlette.websockets

from flama import concurrency, exceptions, http, types, websockets

__all__ = ["HTTPEndpoint", "WebSocketEndpoint"]


class HTTPEndpoint(types.HTTPEndpointProtocol):
    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        """An HTTP endpoint.

        :param scope: ASGI scope.
        :param receive: ASGI receive function.
        :param send: ASGI send function.
        """
        assert scope["type"] == "http"
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        self.state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "path_params": route_scope.get("path_params", {}),
            "route": route,
            "request": http.Request(route_scope, receive=receive),
        }

    def __await__(self) -> t.Generator:
        return self.dispatch().__await__()

    @classmethod
    def allowed_methods(cls) -> set[str]:
        """The list of allowed methods by this endpoint.

        :return: List of allowed methods.
        """
        methods = {
            method for method in http.Method.__members__.keys() if getattr(cls, method.lower(), None) is not None
        }
        if "GET" in methods:
            methods.add("HEAD")
        return methods

    @classmethod
    def allowed_handlers(cls) -> dict[str, t.Callable]:
        """A mapping of handler related to each HTTP method.

        :return: Handlers mapping.
        """
        return {method: getattr(cls, method.lower(), getattr(cls, "get")) for method in cls.allowed_methods()}

    @property
    def handler(self) -> t.Callable:
        """The handler used for dispatching this request.

        :return: Handler.
        """
        handler_name = "get" if self.state["request"].method == "HEAD" else self.state["request"].method.lower()
        h: t.Callable = getattr(self, handler_name)
        return h

    async def dispatch(self) -> None:
        """Dispatch a request."""
        app = self.state["app"]
        handler = await app.injector.inject(self.handler, self.state)
        return await concurrency.run(handler)


class WebSocketEndpoint(types.WebSocketEndpointProtocol):
    encoding: t.Optional[types.Encoding] = None

    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        """A websocket endpoint.

        :param scope: ASGI scope.
        :param receive: ASGI receive function.
        :param send: ASGI send function.
        """
        assert scope["type"] == "websocket"
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        self.state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "path_params": route_scope.get("path_params", {}),
            "route": route,
            "websocket": websockets.WebSocket(route_scope, receive, send),
            "websocket_encoding": self.encoding,
            "websocket_code": None,
            "websocket_message": None,
        }

    def __await__(self) -> t.Generator:
        return self.dispatch().__await__()

    @classmethod
    def allowed_handlers(cls) -> dict[str, t.Callable]:
        """A mapping of handler related to each WS action.

        :return: Handlers mapping.
        """
        return {
            "WEBSOCKET_CONNECT": cls.on_connect,
            "WEBSOCKET_RECEIVE": cls.on_receive,
            "WEBSOCKET_DISCONNECT": cls.on_disconnect,
        }

    async def dispatch(self) -> None:
        """Dispatch a request."""
        app = self.state["app"]
        websocket = self.state["websocket"]

        on_connect = await app.injector.inject(self.on_connect, self.state)
        await on_connect()

        try:
            self.state["websocket_message"] = await websocket.receive()

            while websocket.is_connected:
                on_receive = await app.injector.inject(self.on_receive, self.state)
                await on_receive()
                self.state["websocket_message"] = await websocket.receive()

            self.state["websocket_code"] = types.Code(int(self.state["websocket_message"].get("code", 1000)))
        except starlette.websockets.WebSocketDisconnect as e:
            self.state["websocket_code"] = types.Code(e.code)
            raise exceptions.WebSocketException(e.code, e.reason) from None
        except exceptions.WebSocketException as e:
            self.state["websocket_code"] = types.Code(e.code)
            raise e from None
        except Exception as e:
            self.state["websocket_code"] = types.Code(1011)
            raise e from None
        finally:
            on_disconnect = await app.injector.inject(self.on_disconnect, self.state)
            await on_disconnect()

    async def on_connect(self, websocket: websockets.WebSocket) -> None:
        """Override to handle an incoming websocket connection.

        :param websocket: Websocket.
        """
        await websocket.accept()

    async def on_receive(self, websocket: websockets.WebSocket, data: types.Data) -> None:
        """Override to handle an incoming websocket message.

        :param websocket: Websocket.
        :param data: Received data.
        """
        ...

    async def on_disconnect(self, websocket: websockets.WebSocket, websocket_code: types.Code) -> None:
        """Override to handle a disconnecting websocket.

        :param websocket: Websocket.
        :param websocket_code: Websocket closing code.
        """
        await websocket.close(websocket_code)



import typing as t

import click

from flama.cli.config.config import Config, ExampleConfig

__all__ = ["start", "command"]


@click.command(name="start", context_settings={"auto_envvar_prefix": "FLAMA"})
@click.argument("flama-config", envvar="FLAMA_CONFIG", default="flama.json")
@click.option(
    "--create-config",
    type=click.Choice(["simple", "full"], case_sensitive=False),
    help="Create a config file: 'simple' only includes the host and port of the webserver; 'full' "
    "includes all the details of the webserver.",
)
def command(flama_config: str, create_config: str):
    """Start a Flama Application based on a config file.

    Start a Flama Application with the configuration specified in <FLAMA_CONFIG> file
    (default: flama.json).
    """
    if create_config:
        with open(flama_config, "w") as fs:
            fs.write(ExampleConfig.build(mode=create_config).dumps())
        return

    with open(flama_config) as fs:
        config = Config.load(fs)  # type: ignore[arg-type]

    config.run()


start: t.Callable = command.callback  # type: ignore[assignment]



import enum

from flama import compat

__all__ = ["PaginationType"]


class PaginationType(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    page_number = enum.auto()
    limit_offset = enum.auto()



from http.cookies import SimpleCookie
from urllib.parse import parse_qsl

from flama import http, types
from flama.injection.components import Component, Components

__all__ = [
    "MethodComponent",
    "URLComponent",
    "SchemeComponent",
    "HostComponent",
    "PortComponent",
    "PathComponent",
    "QueryStringComponent",
    "QueryParamsComponent",
    "HeadersComponent",
    "BodyComponent",
    "ASGI_COMPONENTS",
]


class MethodComponent(Component):
    def resolve(self, scope: types.Scope) -> types.Method:
        return types.Method(scope["method"])


class URLComponent(Component):
    def resolve(self, scope: types.Scope) -> types.URL:
        host, port = scope.get("server", ["", None])
        scheme = scope.get("scheme", "")
        path = scope.get("path", "")
        query = scope.get("query_string", b"").decode()

        if (scheme == "http" and port in (80, None)) or (scheme == "https" and port in (443, None)):
            port = None

        if port:
            host += f":{port}"

        if query:
            path += f"?{query}"

        return types.URL(f"{scheme}://{host}{path}")


class SchemeComponent(Component):
    def resolve(self, scope: types.Scope) -> types.Scheme:
        return types.Scheme(scope["scheme"])


class HostComponent(Component):
    def resolve(self, scope: types.Scope) -> types.Host:
        return types.Host(scope["server"][0])


class PortComponent(Component):
    def resolve(self, scope: types.Scope) -> types.Port:
        return types.Port(scope["server"][1])


class PathComponent(Component):
    def resolve(self, scope: types.Scope) -> types.Path:
        return types.Path(scope.get("root_path", "") + scope["path"])


class QueryStringComponent(Component):
    def resolve(self, scope: types.Scope) -> types.QueryString:
        return types.QueryString(scope["query_string"].decode())


class QueryParamsComponent(Component):
    def resolve(self, scope: types.Scope) -> types.QueryParams:
        query_string = scope["query_string"].decode()
        return types.QueryParams(parse_qsl(query_string))


class HeadersComponent(Component):
    def resolve(self, request: http.Request) -> types.Headers:
        return request.headers


class CookiesComponent(Component):
    def resolve(self, headers: types.Headers) -> types.Cookies:
        cookie = SimpleCookie()
        cookie.load(headers.get("cookie", ""))
        return types.Cookies(
            {
                str(name): {**{str(k): str(v) for k, v in morsel.items()}, "value": morsel.value}
                for name, morsel in cookie.items()
            }
        )


class BodyComponent(Component):
    async def resolve(self, receive: types.Receive) -> types.Body:
        body = b""
        while True:
            message = await receive()
            if not message["type"] == "http.request":
                raise Exception(f"Unexpected ASGI message type '{message['type']}'.")
            body += message.get("body", b"")
            if not message.get("more_body", False):
                break

        return types.Body(body)


ASGI_COMPONENTS = Components(
    [
        MethodComponent(),
        URLComponent(),
        SchemeComponent(),
        HostComponent(),
        PortComponent(),
        PathComponent(),
        QueryStringComponent(),
        QueryParamsComponent(),
        HeadersComponent(),
        CookiesComponent(),
        BodyComponent(),
    ]
)



import inspect
import itertools
import typing as t

import marshmallow
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin, resolve_schema_cls

from flama import compat
from flama.injection import Parameter
from flama.schemas._libs.marshmallow.fields import MAPPING, MAPPING_TYPES
from flama.schemas.adapter import Adapter
from flama.schemas.exceptions import SchemaGenerationError, SchemaValidationError
from flama.types import JSONSchema

if t.TYPE_CHECKING:
    from apispec.ext.marshmallow import OpenAPIConverter

__all__ = ["MarshmallowAdapter"]

Schema = marshmallow.Schema
Field = marshmallow.fields.Field


class MarshmallowAdapter(Adapter[Schema, Field]):
    def build_field(
        self,
        name: str,
        type_: type,
        nullable: bool = False,
        required: bool = True,
        default: t.Any = None,
        multiple: bool = False,
        **kwargs,
    ) -> Field:
        field_args = {
            "required": required,
            "allow_none": nullable,
            "metadata": {**kwargs, "title": name},
        }

        if not required:
            field_args["load_default"] = default if default is not Parameter.empty else None

        if multiple:
            return marshmallow.fields.List(
                marshmallow.fields.Nested(type_) if self.is_schema(type_) else MAPPING[type_](),
                **field_args,
            )

        return MAPPING[type_](**field_args)

    def build_schema(
        self,
        *,
        name: t.Optional[str] = None,
        schema: t.Optional[t.Union[Schema, type[Schema]]] = None,
        fields: t.Optional[dict[str, Field]] = None,
        partial: bool = False,
    ) -> type[Schema]:
        fields_ = {**(self.unique_schema(schema)().fields if schema else {}), **(fields or {})}

        if partial:
            for field in fields_:
                fields_[field].required = False
                fields_[field].allow_none = True

        return Schema.from_dict(fields=fields_, name=name or self.DEFAULT_SCHEMA_NAME)  # type: ignore

    def validate(
        self, schema: t.Union[type[Schema], Schema], values: dict[str, t.Any], *, partial: bool = False
    ) -> dict[str, t.Any]:
        try:
            return t.cast(
                dict[str, t.Any],
                self._schema_instance(schema).load(values, unknown=marshmallow.EXCLUDE, partial=partial),
            )
        except marshmallow.ValidationError as exc:
            raise SchemaValidationError(errors=exc.normalized_messages())

    def load(self, schema: t.Union[type[Schema], Schema], value: dict[str, t.Any]) -> Schema:
        return t.cast(Schema, self._schema_instance(schema).load(value))

    def dump(self, schema: t.Union[type[Schema], Schema], value: dict[str, t.Any]) -> dict[str, t.Any]:
        try:
            dump_value = t.cast(dict[str, t.Any], self._schema_instance(schema).dump(value))
        except Exception as exc:
            raise SchemaValidationError(errors=str(exc))

        self.validate(schema, dump_value)

        return dump_value

    def name(self, schema: t.Union[Schema, type[Schema]], *, prefix: t.Optional[str] = None) -> str:
        s = self.unique_schema(schema)
        schema_name = f"{prefix or ''}{s.__qualname__}"
        return schema_name if s.__module__ == "builtins" else f"{s.__module__}.{schema_name}"

    def to_json_schema(self, schema: t.Union[type[Schema], type[Field], Schema, Field]) -> JSONSchema:
        json_schema: dict[str, t.Any]
        try:
            plugin = MarshmallowPlugin(schema_name_resolver=lambda x: t.cast(type, resolve_schema_cls(x)).__name__)
            APISpec("", "", "3.1.0", [plugin])
            converter: "OpenAPIConverter" = t.cast("OpenAPIConverter", plugin.converter)

            if (inspect.isclass(schema) and issubclass(schema, Field)) or isinstance(schema, Field):
                json_schema = converter.field2property(t.cast(marshmallow.fields.Field, schema))
            elif inspect.isclass(schema) and issubclass(schema, Schema):
                json_schema = converter.schema2jsonschema(schema)
            elif isinstance(schema, Schema):
                if getattr(schema, "many", False):
                    json_schema = {
                        "type": "array",
                        "items": converter.schema2jsonschema(schema),
                        "additionalItems": False,
                    }
                else:
                    json_schema = converter.schema2jsonschema(schema)
            else:
                raise SchemaGenerationError

            for property in itertools.chain(json_schema.get("properties", {}).values(), [json_schema]):
                if isinstance(property.get("type"), list):
                    property["anyOf"] = [{"type": x} for x in property["type"]]
                    del property["type"]
        except Exception as e:
            raise SchemaGenerationError from e

        return json_schema

    def unique_schema(self, schema: t.Union[Schema, type[Schema]]) -> type[Schema]:
        if isinstance(schema, Schema):
            return schema.__class__

        return schema

    def _get_field_type(self, field: Field) -> t.Any:
        if isinstance(field, marshmallow.fields.Nested):
            return field.schema

        if isinstance(field, marshmallow.fields.List):
            return self._get_field_type(t.cast(marshmallow.fields.Field, field.inner))

        if isinstance(field, marshmallow.fields.Dict):
            return self._get_field_type(t.cast(marshmallow.fields.Field, field.value_field))

        try:
            return MAPPING_TYPES[field.__class__]
        except KeyError:
            return None

    def schema_fields(
        self, schema: t.Union[Schema, type[Schema]]
    ) -> dict[str, tuple[t.Union[None, type, Schema], Field]]:
        return {
            name: (self._get_field_type(field), field) for name, field in self._schema_instance(schema).fields.items()
        }

    def is_schema(
        self, obj: t.Any
    ) -> compat.TypeGuard[t.Union[Schema, type[Schema]]]:  # PORT: Replace compat when stop supporting 3.9
        return isinstance(obj, Schema) or (inspect.isclass(obj) and issubclass(obj, Schema))

    def is_field(
        self, obj: t.Any
    ) -> compat.TypeGuard[t.Union[Field, type[Field]]]:  # PORT: Replace compat when stop supporting 3.9
        return isinstance(obj, Field) or (inspect.isclass(obj) and issubclass(obj, Field))

    def _schema_instance(self, schema: t.Union[type[Schema], Schema]) -> Schema:
        if inspect.isclass(schema) and issubclass(schema, Schema):
            return schema()
        elif isinstance(schema, Schema):
            return schema
        else:
            raise ValueError("Wrong schema")



import typing as t

from flama import compat

if t.TYPE_CHECKING:
    from flama import endpoints

__all__ = [
    "Scope",
    "Message",
    "Receive",
    "Send",
    "AppClass",
    "AppFunction",
    "App",
    "MiddlewareClass",
    "MiddlewareFunction",
    "Middleware",
    "HTTPHandler",
    "WebSocketHandler",
]

P = compat.ParamSpec("P")  # PORT: Replace compat when stop supporting 3.9
R = t.TypeVar("R", covariant=True)

Scope = t.NewType("Scope", t.MutableMapping[str, t.Any])
Message = t.NewType("Message", t.MutableMapping[str, t.Any])


class Receive(t.Protocol):
    async def __call__(self) -> Message:
        ...


class Send(t.Protocol):
    async def __call__(self, message: Message) -> None:
        ...


# Applications
@t.runtime_checkable
class AppClass(t.Protocol):
    def __call__(self, scope: Scope, receive: Receive, send: Send) -> t.Union[None, t.Awaitable[None]]:
        ...


AppFunction = t.Callable[[Scope, Receive, Send], t.Union[None, t.Awaitable[None]]]
App = t.Union[AppClass, AppFunction]


# Middleware
@t.runtime_checkable
class MiddlewareClass(AppClass, t.Protocol[P, R]):
    def __init__(self, app: App, *args: P.args, **kwargs: P.kwargs):
        ...


MiddlewareFunction = t.Callable[compat.Concatenate[App, P], App]  # PORT: Replace compat when stop supporting 3.9
Middleware = t.Union[type[MiddlewareClass], MiddlewareFunction]

HTTPHandler = t.Union[t.Callable, type["endpoints.HTTPEndpoint"]]
WebSocketHandler = t.Union[t.Callable, type["endpoints.WebSocketEndpoint"]]



import abc
import builtins
import enum
import functools
import inspect
import logging
import typing as t

from flama import compat, concurrency, endpoints, exceptions, http, schemas, types, url, websockets
from flama.injection import Component, Components
from flama.lifespan import Lifespan
from flama.pagination import paginator
from flama.schemas.routing import RouteParametersMixin

if t.TYPE_CHECKING:
    from flama.applications import Flama
    from flama.pagination.types import PaginationType

__all__ = ["Route", "WebSocketRoute", "Mount", "Router"]

logger = logging.getLogger(__name__)


class _EndpointType(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    http = enum.auto()
    websocket = enum.auto()


class Match(enum.Enum):
    none = enum.auto()
    partial = enum.auto()
    full = enum.auto()


class EndpointWrapper:
    type = _EndpointType

    def __init__(
        self,
        handler: t.Union[t.Callable, builtins.type[endpoints.HTTPEndpoint], builtins.type[endpoints.WebSocketEndpoint]],
        endpoint_type: _EndpointType,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
    ):
        """Wraps a function or endpoint into ASGI application.

        :param handler: Function or endpoint.
        :param endpoint_type: Endpoint type, http or websocket.
        :param pagination: Apply a pagination technique.
        """
        if pagination:
            handler = paginator.paginate(pagination, handler)

        self.handler = handler
        functools.update_wrapper(self, handler)
        decorator_select: dict[tuple[_EndpointType, bool], types.App] = {
            (self.type.http, False): self._http_function,
            (self.type.http, True): self._http_endpoint,
            (self.type.websocket, False): self._websocket_function,
            (self.type.websocket, True): self._websocket_endpoint,
        }
        self.call_function = decorator_select[(endpoint_type, inspect.isclass(self.handler))]

    def __get__(self, instance, owner):
        return functools.partial(self.__call__, instance)

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        await concurrency.run(self.call_function, scope, receive, send)

    async def _http_function(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs an HTTP request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        :return: None.
        """
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "path_params": route_scope.get("path_params", {}),
            "route": route,
            "request": http.Request(route_scope, receive=receive),
        }

        try:
            injected_func = await app.injector.inject(self.handler, state)
            response = await concurrency.run(injected_func)
            response = self._build_api_response(self.handler, response)
        except Exception:
            logger.exception("Error performing request")
            raise

        await response(route_scope, receive, send)

    async def _http_endpoint(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs an HTTP request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        :return: None.
        """
        response = await self.handler(scope, receive, send)
        response = self._build_api_response(self.handler, response)

        await response(scope, receive, send)

    async def _websocket_function(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a websocket request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        :return: None.
        """
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "path_params": route_scope.get("path_params", {}),
            "route": route,
            "websocket": websockets.WebSocket(route_scope, receive, send),
            "websocket_encoding": None,
            "websocket_code": None,
            "websocket_message": None,
        }

        injected_func = await app.injector.inject(self.handler, state)
        await injected_func(**route_scope.get("kwargs", {}))

    async def _websocket_endpoint(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a websocket request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        :return: None.
        """
        await self.handler(scope, receive, send)

    def _build_api_response(self, handler: t.Callable, response: t.Union[http.Response, None]) -> http.Response:
        """Build an API response given a handler and the current response.

        It infers the output schema from the handler signature or just wraps the response in a APIResponse object.

        :param handler: The handler in charge of the request.
        :param response: The current response.
        :return: An API response.
        """
        if isinstance(response, (dict, list)):
            try:
                schema = schemas.Schema.from_type(inspect.signature(handler).return_annotation).unique_schema
            except Exception:
                schema = None

            response = http.APIResponse(content=response, schema=schema)
        elif isinstance(response, str):
            response = http.APIResponse(content=response)
        elif response is None:
            response = http.APIResponse(content="")

        return response

    def __eq__(self, other) -> bool:
        return isinstance(other, EndpointWrapper) and self.handler == other.handler


class BaseRoute(abc.ABC, RouteParametersMixin):
    def __init__(
        self,
        path: t.Union[str, url.RegexPath],
        app: types.App,
        *,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        tags: t.Optional[dict[str, t.Any]] = None,
    ):
        """A route definition of a http endpoint.

        :param path: URL path.
        :param app: ASGI application.
        :param name: Route name.
        :param include_in_schema: True if this route must be listed as part of the App schema.
        :param tags: Route tags.
        """
        self.path = url.RegexPath(path)
        self.app = app
        self.endpoint = app.handler if isinstance(app, EndpointWrapper) else app
        self.name = name
        self.include_in_schema = include_in_schema
        self.tags = tags or {}
        super().__init__()

    @abc.abstractmethod
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        ...

    def __eq__(self, other: t.Any) -> bool:
        return (
            isinstance(other, BaseRoute)
            and self.path == other.path
            and self.app == other.app
            and self.name == other.name
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r}, name={(self.name or '')!r})"

    def build(self, app: t.Optional["Flama"] = None) -> None:
        """Build step for routes.

        Just build the parameters' descriptor part of RouteParametersMixin.

        :param app: Flama app.
        """
        if app:
            self.parameters.build(app)

    def endpoint_handlers(self) -> dict[str, t.Callable]:
        """Return a mapping of all possible endpoints of this route.

        Useful to identify all endpoints by HTTP methods.

        :return: Mapping of all endpoints.
        """
        return {}

    async def handle(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request by calling the app of this route.

        :param scope: ASGI scope.
        :param receive: ASGI receive event.
        :param send: ASGI send event.
        """
        await concurrency.run(self.app, scope, receive, send)

    def match(self, scope: types.Scope) -> Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        return Match.full if self.path.match(scope["path"]) else Match.none

    def route_scope(self, scope: types.Scope) -> types.Scope:
        """Build route scope from given scope.

        :param scope: ASGI scope.
        :return: Route scope.
        """
        return types.Scope(
            {
                "endpoint": self.endpoint,
                "path_params": {**dict(scope.get("path_params", {})), **self.path.values(scope["path"])},
            }
        )

    def resolve_url(self, name: str, **params: t.Any) -> url.URL:
        """Builds URL path for given name and params.

        :param name: Route name.
        :param params: Path params.
        :return: URL path.
        """
        if name != self.name:
            raise exceptions.NotFoundException(params=params, name=name)

        try:
            path, remaining_params = self.path.build(**params)
            assert not remaining_params or set(remaining_params.keys()) == {"path"}
        except (ValueError, AssertionError):
            raise exceptions.NotFoundException(params=params, name=name)

        return url.URL(path=path, scheme="http")


class Route(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: t.Union[t.Callable, type[endpoints.HTTPEndpoint]],
        *,
        methods: t.Optional[t.Union[set[str], t.Sequence[str]]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> None:
        """A route definition of a http endpoint.

        :param path: URL path.
        :param endpoint: HTTP endpoint or function.
        :param methods: List of valid HTTP methods.
        :param name: Route name.
        :param include_in_schema: True if this route must be listed as part of the App schema.
        :param pagination: Apply a pagination technique.
        :param tags: Route tags.
        """
        assert self.is_endpoint(endpoint) or (
            not inspect.isclass(endpoint) and callable(endpoint)
        ), "Endpoint must be a callable or an HTTPEndpoint subclass"

        if self.is_endpoint(endpoint):
            self.methods = endpoint.allowed_methods() if methods is None else set(methods)
        else:
            self.methods = {"GET"} if methods is None else set(methods)

        if "GET" in self.methods:
            self.methods.add("HEAD")

        name = endpoint.__name__ if name is None else name

        super().__init__(
            path,
            EndpointWrapper(endpoint, EndpointWrapper.type.http, pagination=pagination),
            name=name,
            include_in_schema=include_in_schema,
            tags=tags,
        )

        self.app: EndpointWrapper

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if scope["type"] == "http":
            await self.handle(types.Scope({**scope, **self.route_scope(scope)}), receive, send)

    def __eq__(self, other: t.Any) -> bool:
        return super().__eq__(other) and isinstance(other, Route) and self.methods == other.methods

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r}, name={self.name!r}, methods={sorted(self.methods)!r})"

    @staticmethod
    def is_endpoint(
        x: t.Union[t.Callable, type[endpoints.HTTPEndpoint]]
    ) -> compat.TypeGuard[type[endpoints.HTTPEndpoint]]:  # PORT: Replace compat when stop supporting 3.9
        return inspect.isclass(x) and issubclass(x, endpoints.HTTPEndpoint)

    def endpoint_handlers(self) -> dict[str, t.Callable]:
        """Return a mapping of all possible endpoints of this route.

        Useful to identify all endpoints by HTTP methods.

        :return: Mapping of all endpoints.
        """
        if self.is_endpoint(self.endpoint):
            return {
                method: handler
                for method, handler in self.endpoint.allowed_handlers().items()
                if method in self.methods
            }

        return {method: self.endpoint for method in self.methods}

    def match(self, scope: types.Scope) -> Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        if scope["type"] != "http":
            return Match.none

        m = super().match(scope)
        if m == Match.none:
            return m

        return Match.full if scope["method"] in self.methods else Match.partial


class WebSocketRoute(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: t.Union[t.Callable, type[endpoints.WebSocketEndpoint]],
        *,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ):
        """A route definition of a websocket endpoint.

        :param path: URL path.
        :param endpoint: Websocket endpoint or function.
        :param name: Route name.
        :param include_in_schema: True if this route must be listed as part of the App schema.
        :param pagination: Apply a pagination technique.
        :param tags: Route tags.
        """

        assert self.is_endpoint(endpoint) or (
            not inspect.isclass(endpoint) and callable(endpoint)
        ), "Endpoint must be a callable or a WebSocketEndpoint subclass"

        name = endpoint.__name__ if name is None else name

        super().__init__(
            path,
            EndpointWrapper(endpoint, EndpointWrapper.type.websocket, pagination=pagination),
            name=name,
            include_in_schema=include_in_schema,
            tags=tags,
        )

        self.app: EndpointWrapper

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if scope["type"] == "websocket":
            await self.handle(types.Scope({**scope, **self.route_scope(scope)}), receive, send)

    def __eq__(self, other: t.Any) -> bool:
        return super().__eq__(other) and isinstance(other, WebSocketRoute)

    @staticmethod
    def is_endpoint(
        x: t.Union[t.Callable, type[endpoints.WebSocketEndpoint]]
    ) -> compat.TypeGuard[type[endpoints.WebSocketEndpoint]]:  # PORT: Replace compat when stop supporting 3.9
        return inspect.isclass(x) and issubclass(x, endpoints.WebSocketEndpoint)

    def endpoint_handlers(self) -> dict[str, t.Callable]:
        """Return a mapping of all possible endpoints of this route.

        Useful to identify all endpoints by HTTP methods.

        :return: Mapping of all endpoints.
        """
        if self.is_endpoint(self.endpoint):
            return self.endpoint.allowed_handlers()

        return {"WEBSOCKET": self.endpoint}

    def match(self, scope: types.Scope) -> Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        if scope["type"] != "websocket":
            return Match.none

        return super().match(scope)


class Mount(BaseRoute):
    def __init__(
        self,
        path: str,
        app: t.Optional[types.App] = None,
        *,
        routes: t.Optional[t.Sequence[BaseRoute]] = None,
        components: t.Optional[t.Sequence[Component]] = None,
        name: t.Optional[str] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ):
        """A mount point for adding a nested ASGI application or a list of routes.

        :param path: URL path.
        :param app: ASGI application.
        :param routes: List of routes.
        :param components: Components registered under this mount point.
        :param name: Mount name.
        :param tags: Mount tags.
        """
        assert app is not None or routes is not None, "Either 'app' or 'routes' must be specified"

        if app is None:
            app = Router(routes=routes, components=components)

        super().__init__(url.RegexPath(path.rstrip("/") + "{path:path}"), app, name=name, tags=tags)

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if scope["type"] in ("http", "websocket") or (
            scope["type"] == "lifespan" and types.is_flama_instance(self.app)
        ):
            await self.handle(types.Scope({**scope, **self.route_scope(scope)}), receive, send)

    def __eq__(self, other: t.Any) -> bool:
        return super().__eq__(other) and isinstance(other, Mount)

    def build(self, app: t.Optional["Flama"] = None) -> None:
        """Build step for routes.

        Just build the parameters' descriptor part of RouteParametersMixin.

        :param app: Flama app.
        """
        if app and types.is_flama_instance(self.app):
            self.app.router.components = Components(self.app.router.components + app.components)

        if root := (self.app if types.is_flama_instance(self.app) else app):
            for route in self.routes:
                route.build(root)

    def match(self, scope: types.Scope) -> Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        if scope["type"] not in ("http", "websocket"):
            return Match.none

        return super().match(scope)

    async def handle(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request by calling the app of this route.

        :param scope: ASGI scope.
        :param receive: ASGI receive event.
        :param send: ASGI send event
        """
        await concurrency.run(self.app, scope, receive, send)

    def route_scope(self, scope: types.Scope) -> types.Scope:
        """Build route scope from given scope.

        It generates an updated scope parameters for the route:
        * app: The app of this mount point. If it's mounting a Flama app, it will replace the app with this one
        * path_params: The matched path parameters of this mount point
        * endpoint: The endpoint of this mount point
        * root_path: The root path of this mount point (if it's mounting a Flama app, it will be empty)
        * path: The remaining path to be matched

        :param scope: ASGI scope.
        :return: Route scope.
        """
        result = {"app": self.app if types.is_flama_instance(self.app) else scope["app"]}

        if "path" in scope:
            path = scope["path"]
            matched_params = self.path.values(path)
            remaining_path = matched_params.pop("path")
            matched_path = path[: -len(remaining_path)]
            result.update(
                {
                    "path_params": {**dict(scope.get("path_params", {})), **matched_params},
                    "endpoint": self.endpoint,
                    "root_path": "" if types.is_flama_instance(self.app) else scope.get("root_path", "") + matched_path,
                    "path": remaining_path,
                }
            )

        return types.Scope(result)

    def resolve_url(self, name: str, **params: t.Any) -> url.URL:
        """Builds URL path for given name and params.

        :param name: Route name.
        :param params: Path params.
        :return: URL path.
        """
        if self.name is not None and name == self.name and "path" in params:
            return super().resolve_url(name, **{**params, "path": params["path"].lstrip("/")})
        elif self.name is None or name.startswith(f"{self.name}:"):
            remaining_name = name if self.name is None else name.split(":", 1)[1]
            path, remaining_params = self.path.build(**{**params, "path": ""})
            if "path" in params:
                remaining_params["path"] = params["path"]
            else:
                del remaining_params["path"]
            for route in self.routes:
                try:
                    route_url = route.resolve_url(remaining_name, **remaining_params)
                    route_url.path = f"{path.rstrip('/')}{route_url.path}"
                    return route_url
                except exceptions.NotFoundException:
                    pass
        raise exceptions.NotFoundException(params=params, name=name)

    @property
    def routes(self) -> list[BaseRoute]:
        """Get all routes registered in this Mount.

        :return: List of routes.
        """
        return getattr(self.app, "routes", [])


class Router:
    def __init__(
        self,
        routes: t.Optional[t.Sequence[BaseRoute]] = None,
        *,
        components: t.Optional[t.Union[t.Sequence["Component"], set["Component"]]] = None,
        lifespan: t.Optional[t.Callable[[t.Optional["Flama"]], t.AsyncContextManager]] = None,
        root: t.Optional["Flama"] = None,
    ):
        """A router for containing all routes and mount points.

        :param routes: Routes part of this router.
        :param components: Components registered in this router.
        :param lifespan: Lifespan function.
        :param root: Flama application.
        """
        self.routes = [] if routes is None else list(routes)
        self.components = Components(components if components else set())
        self.lifespan = Lifespan(lifespan)

        if root:
            self.build(root)

    def __eq__(self, other: t.Any) -> bool:
        return isinstance(other, Router) and self.routes == other.routes

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        logger.debug("Request: %s", str(scope))
        assert scope["type"] in ("http", "websocket", "lifespan")

        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)
            return

        if "router" not in scope:
            scope["router"] = self

        route, route_scope = self.resolve_route(scope)
        await route(route_scope, receive, send)

    def build(self, app: "Flama") -> None:
        """Build step for routes.

        Just build the parameters' descriptor part of RouteParametersMixin.

        :param app: Flama app.
        """
        for route in self.routes:
            route.build(app)

    def add_component(self, component: Component):
        """Register a new component.

        :param component: Component to register.
        """
        self.components = Components(self.components + Components([component]))

    def add_route(
        self,
        path: t.Optional[str] = None,
        endpoint: t.Optional[types.HTTPHandler] = None,
        methods: t.Optional[list[str]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        *,
        route: t.Optional[Route] = None,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> Route:
        """Register a new HTTP route in this router under given path.

        :param path: URL path.
        :param endpoint: HTTP endpoint.
        :param methods: List of valid HTTP methods (only applies for routes).
        :param name: Endpoint or route name.
        :param include_in_schema: True if this route or endpoint should be declared as part of the API schema.
        :param route: HTTP route.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the route or endpoint.
        :return: Route.
        """
        if path is not None and endpoint is not None:
            route = Route(
                path,
                endpoint=endpoint,
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
                pagination=pagination,
                tags=tags,
            )

        assert route is not None, "Either 'path' and 'endpoint' or 'route' variables are needed"

        self.routes.append(route)

        route.build(root)

        return route

    def route(
        self,
        path: str,
        methods: t.Optional[list[str]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        *,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> t.Callable[[types.HTTPHandler], types.HTTPHandler]:
        """Decorator version for registering a new HTTP route in this router under given path.

        :param path: URL path.
        :param methods: List of valid HTTP methods (only applies for routes).
        :param name: Endpoint or route name.
        :param include_in_schema: True if this route or endpoint should be declared as part of the API schema.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the endpoint.
        :return: Decorated route.
        """

        def decorator(func: types.HTTPHandler) -> types.HTTPHandler:
            self.add_route(
                path,
                func,
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
                root=root,
                pagination=pagination,
                tags=tags,
            )
            return func

        return decorator

    def add_websocket_route(
        self,
        path: t.Optional[str] = None,
        endpoint: t.Optional[types.WebSocketHandler] = None,
        name: t.Optional[str] = None,
        *,
        route: t.Optional[WebSocketRoute] = None,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> WebSocketRoute:
        """Register a new websocket route in this router under given path.

        :param path: URL path.
        :param endpoint: Websocket function or endpoint.
        :param name: Websocket route name.
        :param route: Specific route class.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the websocket route.
        :return: Websocket route.
        """
        if path is not None and endpoint is not None:
            route = WebSocketRoute(path, endpoint=endpoint, name=name, pagination=pagination, tags=tags)

        assert route is not None, "Either 'path' and 'endpoint' or 'route' variables are needed"

        self.routes.append(route)

        route.build(root)

        return route

    def websocket_route(
        self,
        path: str,
        name: t.Optional[str] = None,
        *,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> t.Callable[[types.WebSocketHandler], types.WebSocketHandler]:
        """Decorator version for registering a new websocket route in this router under given path.

        :param path: URL path.
        :param name: Websocket route name.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the websocket route.
        :return: Decorated websocket route.
        """

        def decorator(func: types.WebSocketHandler) -> types.WebSocketHandler:
            self.add_websocket_route(path, func, name=name, root=root, pagination=pagination, tags=tags)
            return func

        return decorator

    def mount(
        self,
        path: t.Optional[str] = None,
        app: t.Optional[types.App] = None,
        name: t.Optional[str] = None,
        *,
        mount: t.Optional[Mount] = None,
        root: t.Optional["Flama"] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> Mount:
        """Register a new mount point containing an ASGI app in this router under given path.

        :param path: URL path.
        :param app: ASGI app to mount.
        :param name: Route name.
        :param mount: Mount.
        :param root: Flama application.
        :param tags: Tags to add to the mount.
        :return: Mount.
        """
        if path is not None and app is not None:
            mount = Mount(path, app=app, name=name, tags=tags)

        assert mount is not None, "Either 'path' and 'app' or 'mount' variables are needed"

        self.routes.append(mount)

        mount.build(root)

        return mount

    def resolve_route(self, scope: types.Scope) -> tuple[BaseRoute, types.Scope]:
        """Look for a route that matches given ASGI scope.

        :param scope: ASGI scope.
        :return: Route and its scope.
        """
        partial = None
        partial_allowed_methods: set[str] = set()

        for route in self.routes:
            m = route.match(scope)
            if m == Match.full:
                route_scope = types.Scope({**scope, **route.route_scope(scope)})

                if isinstance(route, Mount):
                    try:
                        return route.app.resolve_route(route_scope)  # type: ignore[no-any-return,union-attr]
                    except AttributeError:
                        ...

                return route, route_scope
            elif m == Match.partial:
                partial = route
                if isinstance(route, Route):
                    partial_allowed_methods |= route.methods

        if partial:
            route_scope = types.Scope({**scope, **partial.route_scope(scope)})
            raise exceptions.MethodNotAllowedException(
                route_scope.get("root_path", "") + route_scope["path"],
                route_scope["method"],
                partial_allowed_methods,
                params=route_scope.get("path_params"),
            )

        raise exceptions.NotFoundException(
            path=scope.get("root_path", "") + scope["path"], params=scope.get("path_params")
        )

    def resolve_url(self, name: str, **path_params: t.Any) -> url.URL:
        """Look for a route URL given the route name and path params.

        :param name: Route name.
        :param path_params: Path params.
        :return: Route URL.
        """
        for route in self.routes:
            try:
                return route.resolve_url(name, **path_params)
            except exceptions.NotFoundException:
                pass
        raise exceptions.NotFoundException(params=path_params, name=name)



import functools
import logging
import threading
import typing as t

from flama import asgi, exceptions, http, injection, types, url, validation, websockets
from flama.ddd.components import WorkerComponent
from flama.events import Events
from flama.middleware import MiddlewareStack
from flama.models.modules import ModelsModule
from flama.modules import Modules
from flama.pagination import paginator
from flama.resources import ResourcesModule
from flama.routing import BaseRoute, Router
from flama.schemas.modules import SchemaModule

try:
    from flama.resources.workers import FlamaWorker
except exceptions.DependencyNotInstalled:
    FlamaWorker = None

if t.TYPE_CHECKING:
    from flama.middleware import Middleware
    from flama.modules import Module
    from flama.pagination.types import PaginationType
    from flama.routing import Mount, Route, WebSocketRoute

__all__ = ["Flama"]

logger = logging.getLogger(__name__)


class Flama:
    def __init__(
        self,
        routes: t.Optional[t.Sequence[t.Union["BaseRoute", "Mount"]]] = None,
        components: t.Optional[t.Union[t.Sequence[injection.Component], set[injection.Component]]] = None,
        modules: t.Optional[t.Union[t.Sequence["Module"], set["Module"]]] = None,
        middleware: t.Optional[t.Sequence["Middleware"]] = None,
        debug: bool = False,
        events: t.Optional[t.Union[dict[str, list[t.Callable[..., t.Coroutine[t.Any, t.Any, None]]]], Events]] = None,
        lifespan: t.Optional[t.Callable[[t.Optional["Flama"]], t.AsyncContextManager]] = None,
        title: str = "Flama",
        version: str = "0.1.0",
        description: str = "Firing up with the flame",
        schema: t.Optional[str] = "/schema/",
        docs: t.Optional[str] = "/docs/",
        schema_library: t.Optional[str] = None,
    ) -> None:
        """Flama application.

        :param routes: Routes part of this application.
        :param components: Components registered in this application.
        :param modules: Modules for extending the application.
        :param middleware: List of middlewares to include in call stack.
        :param debug: Debug mode.
        :param events: Handlers that will be triggered after certain events.
        :param lifespan: Lifespan function.
        :param title: API title.
        :param version: API version.
        :param description: API description.
        :param schema: OpenAPI schema endpoint path.
        :param docs: Docs endpoint path.
        :param schema_library: Schema library to use.
        """
        self._debug = debug
        self._status = types.AppStatus.NOT_STARTED
        self._shutdown = False

        # Create Dependency Injector
        self._injector = injection.Injector(
            context_types={
                "scope": types.Scope,
                "receive": types.Receive,
                "send": types.Send,
                "exc": Exception,
                "app": Flama,
                "path_params": types.PathParams,
                "route": BaseRoute,
                "request": http.Request,
                "response": http.Response,
                "websocket": websockets.WebSocket,
                "websocket_message": types.Message,
                "websocket_encoding": types.Encoding,
                "websocket_code": types.Code,
            }
        )

        # Initialise components
        default_components = []

        # Create worker
        if (worker := FlamaWorker() if FlamaWorker else None) and WorkerComponent:
            default_components.append(WorkerComponent(worker=worker))

        # Initialise modules
        default_modules = [
            ResourcesModule(worker=worker),
            SchemaModule(title, version, description, schema=schema, docs=docs),
            ModelsModule(),
        ]
        self.modules = Modules(app=self, modules={*default_modules, *(modules or [])})

        # Initialize router
        self.app = self.router = Router(
            routes=routes, components=[*default_components, *(components or [])], lifespan=lifespan
        )

        # Build middleware stack
        self.middleware = MiddlewareStack(app=self, middleware=middleware or [], debug=debug)

        # Setup schema library
        self.schema.schema_library = schema_library

        # Add schema routes
        self.schema.add_routes()

        # Build events register including module events
        self.events = events if isinstance(events, Events) else Events.build(**(events or {}))
        self.events.startup += [m.on_startup for m in self.modules.values()]
        self.events.shutdown += [m.on_shutdown for m in self.modules.values()]

        # Reference to paginator from within app
        self.paginator = paginator

        # Build router to propagate root application
        self.router.build(self)

    def __getattr__(self, item: str) -> t.Any:
        """Retrieve a module by its name.

        :param item: Module name.
        :return: Module.
        """
        try:
            return self.modules.__getitem__(item)
        except KeyError:
            return None  # type: ignore[return-value]

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Perform a request.

        :param scope: ASGI scope.
        :param receive: ASGI receive event.
        :param send: ASGI send event.
        """
        if scope["type"] != "lifespan" and self.status in (types.AppStatus.NOT_STARTED, types.AppStatus.STARTING):
            raise exceptions.ApplicationError("Application is not ready to process requests yet.")

        if scope["type"] != "lifespan" and self.status in (types.AppStatus.SHUT_DOWN, types.AppStatus.SHUTTING_DOWN):
            raise exceptions.ApplicationError("Application is already shut down.")

        scope["app"] = self
        scope.setdefault("root_app", self)
        await self.middleware(scope, receive, send)

    @property
    def status(self) -> types.AppStatus:
        return self._status

    @status.setter
    def status(self, s: types.AppStatus) -> None:
        logger.debug("Transitioning %s from %s to %s", self, self._status, s)

        with threading.Lock():
            self._status = s

    @property
    def components(self) -> injection.Components:
        """Components register.

        :return: Components register.
        """
        return self.router.components

    def add_component(self, component: injection.Component):
        """Add a new component to the register.

        :param component: Component to include.
        """
        self.router.add_component(component)
        self.router.build(self)

    @property
    def routes(self) -> list["BaseRoute"]:
        """List of registered routes.

        :return: Routes.
        """
        return self.router.routes

    def add_route(
        self,
        path: t.Optional[str] = None,
        endpoint: t.Optional[types.HTTPHandler] = None,
        methods: t.Optional[list[str]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        route: t.Optional["Route"] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> "Route":
        """Register a new HTTP route or endpoint under given path.

        :param path: URL path.
        :param endpoint: HTTP endpoint.
        :param methods: List of valid HTTP methods (only applies for routes).
        :param name: Endpoint or route name.
        :param include_in_schema: True if this route or endpoint should be declared as part of the API schema.
        :param route: HTTP route.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the route.
        """
        return self.router.add_route(
            path,
            endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
            route=route,
            root=self,
            pagination=pagination,
            tags=tags,
        )

    def route(
        self,
        path: str,
        methods: t.Optional[list[str]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> t.Callable[[types.HTTPHandler], types.HTTPHandler]:
        """Decorator version for registering a new HTTP route in this router under given path.

        :param path: URL path.
        :param methods: List of valid HTTP methods (only applies for routes).
        :param name: Endpoint or route name.
        :param include_in_schema: True if this route or endpoint should be declared as part of the API schema.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the route.
        :return: Decorated route.
        """
        return self.router.route(
            path,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
            root=self,
            pagination=pagination,
            tags=tags,
        )

    def add_websocket_route(
        self,
        path: t.Optional[str] = None,
        endpoint: t.Optional[types.WebSocketHandler] = None,
        name: t.Optional[str] = None,
        route: t.Optional["WebSocketRoute"] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> "WebSocketRoute":
        """Register a new websocket route or endpoint under given path.

        :param path: URL path.
        :param endpoint: Websocket endpoint.
        :param name: Endpoint or route name.
        :param route: Websocket route.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the websocket route.
        """
        return self.router.add_websocket_route(
            path, endpoint, name=name, route=route, root=self, pagination=pagination, tags=tags
        )

    def websocket_route(
        self,
        path: str,
        name: t.Optional[str] = None,
        pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> t.Callable[[types.WebSocketHandler], types.WebSocketHandler]:
        """Decorator version for registering a new websocket route in this router under given path.

        :param path: URL path.
        :param name: Websocket route name.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the websocket route.
        :return: Decorated route.
        """
        return self.router.websocket_route(path, name=name, root=self, pagination=pagination, tags=tags)

    def mount(
        self,
        path: t.Optional[str] = None,
        app: t.Optional[types.App] = None,
        name: t.Optional[str] = None,
        mount: t.Optional["Mount"] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> "Mount":
        """Register a new mount point containing an ASGI app in this router under given path.

        :param path: URL path.
        :param app: ASGI app to mount.
        :param name: Application name.
        :param mount: Mount.
        :param tags: Tags to add to the mount.
        :return: Mount.
        """
        return self.router.mount(path, app, name=name, mount=mount, root=self, tags=tags)

    @property
    def injector(self) -> injection.Injector:
        """Components dependency injector.

        :return: Injector instance.
        """
        components = injection.Components(self.components + asgi.ASGI_COMPONENTS + validation.VALIDATION_COMPONENTS)
        if self._injector.components != components:
            self._injector.components = components
        return self._injector

    def add_event_handler(self, event: str, func: t.Callable) -> None:
        """Register a new event handler.

        :param event: Event type.
        :param func: Event handler.
        """
        self.events.register(event, func)

    def on_event(self, event: str) -> t.Callable:
        """Decorator version for registering a new event handler.

        :param event: Event type.
        :return: Decorated handler.
        """

        def decorator(func: t.Callable) -> t.Callable:
            self.add_event_handler(event, func)
            return func

        return decorator

    def add_exception_handler(self, exc_class_or_status_code: t.Union[int, type[Exception]], handler: t.Callable):
        """Add a new exception handler for given status code or exception class.

        :param exc_class_or_status_code: Status code or exception class.
        :param handler: Exception handler.
        """
        self.middleware.add_exception_handler(exc_class_or_status_code, handler)

    def add_middleware(self, middleware: "Middleware"):
        """Add a new middleware to the stack.

        :param middleware: Middleware instance.
        """
        self.middleware.add_middleware(middleware)

    def resolve_url(self, name: str, **path_params: t.Any) -> url.URL:
        """Look for a route URL given the route name and path params.

        :param name: Route name.
        :param path_params: Path params.
        :return: Route URL.
        """
        return self.router.resolve_url(name, **path_params)

    def resolve_route(self, scope: types.Scope) -> tuple[BaseRoute, types.Scope]:
        """Look for a route that matches given ASGI scope.

        :param scope: ASGI scope.
        :return: Route and its scope.
        """
        return self.router.resolve_route(scope)

    get = functools.partialmethod(route, methods=["GET"])
    head = functools.partialmethod(route, methods=["HEAD"])
    post = functools.partialmethod(route, methods=["POST"])
    put = functools.partialmethod(route, methods=["PUT"])
    delete = functools.partialmethod(route, methods=["DELETE"])
    connect = functools.partialmethod(route, methods=["CONNECT"])
    options = functools.partialmethod(route, methods=["OPTIONS"])
    trace = functools.partialmethod(route, methods=["TRACE"])
    patch = functools.partialmethod(route, methods=["PATCH"])
    add_get = functools.partialmethod(add_route, methods=["GET"])
    add_head = functools.partialmethod(add_route, methods=["HEAD"])
    add_post = functools.partialmethod(add_route, methods=["POST"])
    add_put = functools.partialmethod(add_route, methods=["PUT"])
    add_delete = functools.partialmethod(add_route, methods=["DELETE"])
    add_connect = functools.partialmethod(add_route, methods=["CONNECT"])
    add_options = functools.partialmethod(add_route, methods=["OPTIONS"])
    add_trace = functools.partialmethod(add_route, methods=["TRACE"])
    add_patch = functools.partialmethod(add_route, methods=["PATCH"])



import enum
import functools
import typing as t

import starlette.background

from flama import compat, concurrency

__all__ = ["BackgroundTask", "BackgroundTasks", "Concurrency", "BackgroundThreadTask", "BackgroundProcessTask"]

P = compat.ParamSpec("P")  # PORT: Replace compat when stop supporting 3.9


class task_wrapper:
    def __init__(self, target: t.Callable[P, t.Union[None, t.Awaitable[None]]]):
        self.target = target
        functools.update_wrapper(self, target)

    async def __call__(self, *args, **kwargs):
        await concurrency.run(self.target, *args, **kwargs)


class Concurrency(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    thread = enum.auto()
    process = enum.auto()


class BackgroundTask(starlette.background.BackgroundTask):
    def __init__(
        self,
        concurrency: t.Union[Concurrency, str],
        func: t.Callable[P, t.Union[None, t.Awaitable[None]]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.func = task_wrapper(func)
        self.args = args
        self.kwargs = kwargs
        self.concurrency = Concurrency[concurrency]

    async def __call__(self):
        if self.concurrency == Concurrency.process:
            concurrency.AsyncProcess(target=self.func, args=self.args, kwargs=self.kwargs).start()
        else:
            await self.func(*self.args, **self.kwargs)


class BackgroundTasks(BackgroundTask):
    def __init__(self, tasks: t.Optional[t.Sequence[BackgroundTask]] = None):
        self.tasks = list(tasks) if tasks else []

    def add_task(
        self, concurrency: t.Union[Concurrency, str], func: t.Callable[P, t.Any], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.tasks.append(BackgroundTask(concurrency, func, *args, **kwargs))

    async def __call__(self) -> None:
        for task in self.tasks:
            await task()


class BackgroundThreadTask(BackgroundTask):
    def __init__(self, func: t.Callable[P, t.Any], *args: P.args, **kwargs: P.kwargs):
        super().__init__(Concurrency.thread, func, *args, **kwargs)


class BackgroundProcessTask(BackgroundTask):
    def __init__(self, func: t.Callable[P, t.Any], *args: P.args, **kwargs: P.kwargs):
        super().__init__(Concurrency.process, func, *args, **kwargs)
